from __future__ import annotations

import sys
import typing

if sys.version_info >= (3, 11):
    from typing import Self
elif typing.TYPE_CHECKING:
    from typing_extensions import Self
else:
    Self = object

__all__ = ["Self"]


def __dir__() -> list[str]:
    return __all__



from typing import Any

from typing_extensions import Self

class _BaseStorage:
    def __init__(self) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __copy__(self) -> Self: ...
    def __deepcopy__(self, memo: Any) -> Self: ...

class int64(_BaseStorage): ...
class double(_BaseStorage): ...
class atomic_int64(_BaseStorage): ...
class unlimited(_BaseStorage): ...
class weight(_BaseStorage): ...
class mean(_BaseStorage): ...
class weighted_mean(_BaseStorage): ...



"""
This test module does simple smoke checks to make sure boost_histogram compiled
correctly. Run with `python -m boost_histogram.test`.
"""

from __future__ import annotations

import pickle
import unittest

import boost_histogram as bh


class TestBoostHistogram(unittest.TestCase):
    def test_1d_histogram(self) -> None:
        h = bh.Histogram(bh.axis.Regular(10, 0, 1))
        h.fill([0.15, 0.25, 0.25, 0.35])
        self.assertEqual(h[0], 0)
        self.assertEqual(h[1], 1)
        self.assertEqual(h[2], 2)
        self.assertEqual(h[3], 1)
        self.assertEqual(h[4], 0)

    def test_2d_histogram(self) -> None:
        h = bh.Histogram(
            bh.axis.Integer(4, 10),
            bh.axis.StrCategory(["a", "b", "c"], overflow=False),
            storage=bh.storage.Weight(),
        )
        h.fill([5, 5, 5, 7], ["a", "b", "b", "c"])

        val = h[bh.loc(5), bh.loc("a")]
        assert isinstance(val, bh.accumulators.WeightedSum)
        self.assertEqual(val.value, 1)

        val = h[bh.loc(5), bh.loc("b")]
        assert isinstance(val, bh.accumulators.WeightedSum)
        self.assertEqual(val.value, 2)

        val = h[bh.loc(7), bh.loc("c")]
        assert isinstance(val, bh.accumulators.WeightedSum)
        self.assertEqual(val.value, 1)

        val = h[bh.loc(8), bh.loc("c")]
        assert isinstance(val, bh.accumulators.WeightedSum)
        self.assertEqual(val.value, 0)

    def test_pickle(self) -> None:
        h = bh.Histogram(
            bh.axis.Variable([0, 0.3, 0.4, 0.5, 1.0]), storage=bh.storage.Mean()
        )
        h.fill([0.15, 0.25, 0.25, 0.35], sample=[1, 2, 3, 4])
        pickled = pickle.dumps(h)
        unpickled = pickle.loads(pickled)
        self.assertEqual(h, unpickled)

    def test_sqrt_transform(self) -> None:
        a = bh.axis.Regular(10, 0, 10, transform=bh.axis.transform.sqrt)
        # Edges: 0. ,  0.1,  0.4,  0.9,  1.6,  2.5,  3.6,  4.9,  6.4,  8.1, 10.

        self.assertEqual(a.index(-100), 10)  # Always in overflow bin
        self.assertEqual(a.index(-1), 10)  # When transform is invalid
        self.assertEqual(a.index(0), 0)
        self.assertEqual(a.index(0.15), 1)
        self.assertEqual(a.index(0.5), 2)
        self.assertEqual(a.index(1), 3)
        self.assertEqual(a.index(1.7), 4)
        self.assertEqual(a.index(9), 9)
        self.assertEqual(a.index(11), 10)
        self.assertEqual(a.index(1000), 10)

        val = a.bin(0)
        assert isinstance(val, tuple)
        self.assertAlmostEqual(val[0], 0.0)

        val = a.bin(1)
        assert isinstance(val, tuple)
        self.assertAlmostEqual(val[0], 0.1)

        val = a.bin(2)
        assert isinstance(val, tuple)
        self.assertAlmostEqual(val[0], 0.4)


if __name__ == "__main__":
    unittest.main("boost_histogram.test", warnings="error")



from __future__ import annotations

from typing import ClassVar

import boost_histogram

from ._core import accumulators  # pylint: disable=no-name-in-module
from ._core import storage as store  # pylint: disable=no-name-in-module

__all__ = [
    "AtomicInt64",
    "Double",
    "Int64",
    "Mean",
    "Storage",
    "Unlimited",
    "Weight",
    "WeightedMean",
]


def __dir__() -> list[str]:
    return __all__


# Simple mixin to provide a common base class for types
class Storage:
    _family: object

    def __init_subclass__(cls, *, family: object) -> None:
        super().__init_subclass__()
        cls._family = family

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    accumulator: ClassVar[
        (
            type[
                int
                | float
                | accumulators.WeightedMean
                | accumulators.WeightedSum
                | accumulators.Mean
            ]
        )
    ]


class Int64(store.int64, Storage, family=boost_histogram):
    accumulator = int


class Double(store.double, Storage, family=boost_histogram):
    accumulator = float


class AtomicInt64(store.atomic_int64, Storage, family=boost_histogram):
    accumulator = int


class Unlimited(store.unlimited, Storage, family=boost_histogram):
    accumulator = float


class Weight(store.weight, Storage, family=boost_histogram):
    accumulator = accumulators.WeightedSum


class Mean(store.mean, Storage, family=boost_histogram):
    accumulator = accumulators.Mean


class WeightedMean(store.weighted_mean, Storage, family=boost_histogram):
    accumulator = accumulators.WeightedMean



from __future__ import annotations

import functools
from typing import Any

import numpy as np

from .. import storage

__all__ = ["_data_from_dict", "_storage_from_dict", "_storage_to_dict"]


def __dir__() -> list[str]:
    return __all__


@functools.singledispatch
def _storage_to_dict(_storage: Any, /, data: Any) -> dict[str, Any]:  # noqa: ARG001
    """Convert a storage to a dictionary."""
    msg = f"Unsupported storage type: {_storage}"
    raise TypeError(msg)


@_storage_to_dict.register(storage.Int64)
def _(_storage: storage.Int64, /, data: Any) -> dict[str, Any]:
    return {"type": "int", "values": data}


@_storage_to_dict.register(storage.Double)
def _(_storage: storage.Double, /, data: Any) -> dict[str, Any]:
    return {"type": "double", "values": data}


@_storage_to_dict.register(storage.AtomicInt64)
@_storage_to_dict.register(storage.Unlimited)
def _(
    storage_: storage.AtomicInt64 | storage.Unlimited,
    /,
    data: Any,
) -> dict[str, Any]:
    return {
        "writer_info": {"boost-histogram": {"orig_type": type(storage_).__name__}},
        "type": "int" if np.issubdtype(data.dtype, np.integer) else "double",
        "values": data,
    }


@_storage_to_dict.register(storage.Weight)
def _(_storage: storage.Weight, /, data: Any) -> dict[str, Any]:
    return {
        "type": "weighted",
        "values": data.value,
        "variances": data.variance,
    }


@_storage_to_dict.register(storage.Mean)
def _(_storage: storage.Mean, /, data: Any) -> dict[str, Any]:
    return {
        "type": "mean",
        "counts": data.count,
        "values": data.value,
        "variances": data.variance,
    }


@_storage_to_dict.register(storage.WeightedMean)
def _(_storage: storage.WeightedMean, /, data: Any) -> dict[str, Any]:
    return {
        "type": "weighted_mean",
        "sum_of_weights": data.sum_of_weights,
        "sum_of_weights_squared": data.sum_of_weights_squared,
        "values": data.value,
        "variances": data.variance,
    }


def _storage_from_dict(data: dict[str, Any], /) -> storage.Storage:
    """Convert a dictionary to a storage object."""
    # If loading a boost-histogram, we can load the exact original type
    orig_type = (
        data.get("writer_info", {}).get("boost-histogram", {}).get("orig_type", "")
    )
    if orig_type == "AtomicInt64":
        return storage.AtomicInt64()
    if orig_type == "Unlimited":
        return storage.Unlimited()

    storage_type = data["type"]
    if storage_type == "int":
        return storage.Int64()
    if storage_type == "double":
        return storage.Double()
    if storage_type == "weighted":
        return storage.Weight()
    if storage_type == "mean":
        return storage.Mean()
    if storage_type == "weighted_mean":
        return storage.WeightedMean()

    raise TypeError(f"Unsupported storage type: {storage_type}")


def _data_from_dict(data: dict[str, Any], /) -> np.typing.NDArray[Any]:
    """Convert a dictionary to data."""
    storage_type = data["type"]

    if storage_type in {"int", "double"}:
        return data["values"]
    if storage_type == "weighted":
        return np.stack([data["values"], data["variances"]], axis=-1)
    if storage_type == "mean":
        return np.stack([data["counts"], data["values"], data["variances"]], axis=-1)
    if storage_type == "weighted_mean":
        return np.stack(
            [
                data["sum_of_weights"],
                data["sum_of_weights_squared"],
                data["values"],
                data["variances"],
            ],
            axis=-1,
        )

    raise TypeError(f"Unsupported storage type: {storage_type}")



from __future__ import annotations

from collections.abc import Mapping, MutableMapping
from typing import Any, Callable, ClassVar, Literal

import numpy as np

from .accumulators import Mean, WeightedMean, WeightedSum
from .typing import ArrayLike, StrIndex, Ufunc

UFMethod = Literal["__call__", "reduce", "reduceat", "accumulate", "outer", "at"]


class View(np.ndarray):
    __slots__ = ()
    _FIELDS: ClassVar[tuple[str, ...]]
    _PARENT: type[WeightedSum | WeightedMean | Mean]

    def __getitem__(self, ind: StrIndex) -> np.typing.NDArray[Any]:  # type: ignore[override]
        sliced = super().__getitem__(ind)

        # If the shape is empty, return the parent type
        if not sliced.shape:
            return self._PARENT._make(*sliced)

        # If the dtype has changed, return a normal array (no longer a record)
        if sliced.dtype != self.dtype:
            return np.asarray(sliced)

        # Otherwise, no change, return the same View type
        return sliced

    def __repr__(self) -> str:
        # NumPy starts the ndarray class name with "array", so we replace it
        # with our class name
        return f"{self.__class__.__name__}(\n      " + repr(self.view(np.ndarray))[6:]

    def __str__(self) -> str:
        my_fields = ", ".join(self._FIELDS)
        return f"{self.__class__.__name__}: ({my_fields})\n{self.view(np.ndarray)}"

    def __setitem__(self, ind: StrIndex, value: ArrayLike) -> None:  # type: ignore[override]
        # `.value` really is ["value"] for an record array
        if isinstance(ind, str):
            super().__setitem__(ind, value)
            return

        current_ndim = super().__getitem__(ind).ndim

        array: np.typing.NDArray[Any] = np.asarray(value)
        msg = "Needs matching ndarray or n+1 dim array"
        if array.ndim == current_ndim + 1:
            if len(self._FIELDS) == array.shape[-1]:
                self.__setitem__(ind, self._PARENT._array(*np.moveaxis(array, -1, 0)))
                return
            msg += f", final dimension should be {len(self._FIELDS)} for this storage, got {array.shape[-1]} instead"
            raise ValueError(msg)
        if self.dtype == array.dtype:
            super().__setitem__(ind, array)
            return

        msg += f", {current_ndim}D {self.dtype} or {current_ndim + 1}D required, got {array.ndim}D {array.dtype}"
        raise ValueError(msg)


def make_getitem_property(name: str) -> property:
    def fget(self: Mapping[str, Any]) -> Any:
        return self[name]

    def fset(self: MutableMapping[str, Any], value: Any) -> None:
        self[name] = value

    return property(fget, fset)


def fields(*names: str) -> Callable[[type[object]], type[object]]:
    """
    This decorator adds the name to the _FIELDS
    class property (for printing in reprs), and
    adds a property that looks like this:

    @property
    def name(self):
        return self["name"]
    @name.setter
    def name(self, value):
        self["name"] = value
    """

    def injector(cls: type[object]) -> type[object]:
        if hasattr(cls, "_FIELDS"):
            raise RuntimeError(
                f"{cls.__name__} already has had a fields decorator applied"
            )
        my_fields = []
        for name in names:
            my_fields.append(name)
            setattr(cls, name, make_getitem_property(name))
            cls._FIELDS = tuple(my_fields)  # type: ignore[attr-defined]

        return cls

    return injector


@fields("value", "variance")
class WeightedSumView(View):
    __slots__ = ()
    _PARENT = WeightedSum

    value: np.typing.NDArray[Any]
    variance: np.typing.NDArray[Any]

    # Could be implemented on master View
    def __array_ufunc__(
        self, ufunc: Ufunc, method: UFMethod, *inputs: Any, **kwargs: Any
    ) -> np.typing.NDArray[Any]:
        # Avoid infinite recursion
        raw_inputs = [np.asarray(x) for x in inputs]

        # Support unary + and -
        if (
            method == "__call__"
            and len(raw_inputs) == 1
            and ufunc in {np.negative, np.positive}
        ):
            (result,) = (
                kwargs.pop("out")
                if "out" in kwargs
                else [np.empty(self.shape, self.dtype)]
            )

            ufunc(raw_inputs[0]["value"], out=result["value"], **kwargs)
            result["variance"] = raw_inputs[0]["variance"]
            return result.view(self.__class__)

        if method == "__call__" and len(raw_inputs) == 2:
            (result,) = (
                kwargs.pop("out")
                if "out" in kwargs
                else [np.empty(self.shape, self.dtype)]
            )

            # Addition of two views
            if raw_inputs[0].dtype == raw_inputs[1].dtype:
                if ufunc in {np.add, np.subtract}:
                    ufunc(
                        raw_inputs[0]["value"],
                        raw_inputs[1]["value"],
                        out=result["value"],
                        **kwargs,
                    )
                    np.add(
                        raw_inputs[0]["variance"],
                        raw_inputs[1]["variance"],
                        out=result["variance"],
                        **kwargs,
                    )
                    return result.view(self.__class__)

                # If unsupported, just pass through (will return not implemented)
                # pylint: disable-next=no-member
                return super().__array_ufunc__(ufunc, method, *raw_inputs, **kwargs)

            # View with normal value or array
            if ufunc in {np.add, np.subtract}:
                if self.dtype == raw_inputs[0].dtype:
                    ufunc(
                        raw_inputs[0]["value"],
                        raw_inputs[1],
                        out=result["value"],
                        **kwargs,
                    )
                    np.add(
                        raw_inputs[0]["variance"],
                        raw_inputs[1] ** 2,
                        out=result["variance"],
                        **kwargs,
                    )
                else:
                    ufunc(
                        raw_inputs[0],
                        raw_inputs[1]["value"],
                        out=result["value"],
                        **kwargs,
                    )
                    np.add(
                        raw_inputs[0] ** 2,
                        raw_inputs[1]["variance"],
                        out=result["variance"],
                        **kwargs,
                    )
                return result.view(self.__class__)

            if ufunc in {np.multiply, np.divide, np.true_divide, np.floor_divide}:
                if self.dtype == raw_inputs[0].dtype:
                    ufunc(
                        raw_inputs[0]["value"],
                        raw_inputs[1],
                        out=result["value"],
                        **kwargs,
                    )
                    ufunc(
                        raw_inputs[0]["variance"],
                        raw_inputs[1] ** 2,
                        out=result["variance"],
                        **kwargs,
                    )
                else:
                    ufunc(
                        raw_inputs[0],
                        raw_inputs[1]["value"],
                        out=result["value"],
                        **kwargs,
                    )
                    ufunc(
                        raw_inputs[0] ** 2,
                        raw_inputs[1]["variance"],
                        out=result["variance"],
                        **kwargs,
                    )

                return result.view(self.__class__)

        # ufuncs that are allowed to reduce
        if ufunc in {np.add} and method == "reduce" and len(raw_inputs) == 1:
            results = (ufunc.reduce(self[field], **kwargs) for field in self._FIELDS)
            return self._PARENT._make(*results)

        # ufuncs that are allowed to accumulate
        if ufunc in {np.add} and method == "accumulate" and len(raw_inputs) == 1:
            (result,) = (
                kwargs.pop("out")
                if "out" in kwargs
                else [np.empty(self.shape, self.dtype)]
            )
            for field in self._FIELDS:
                ufunc.accumulate(self[field], out=result[field], **kwargs)
            return result.view(self.__class__)

        # If unsupported, just pass through (will return NotImplemented or things like == will work but not return subclasses)
        # pylint: disable-next=no-member
        return super().__array_ufunc__(ufunc, method, *raw_inputs, **kwargs)


@fields(
    "sum_of_weights",
    "sum_of_weights_squared",
    "value",
    "_sum_of_weighted_deltas_squared",
)
class WeightedMeanView(View):
    __slots__ = ()
    _PARENT = WeightedMean

    sum_of_weights: np.typing.NDArray[Any]
    sum_of_weights_squared: np.typing.NDArray[Any]
    value: np.typing.NDArray[Any]
    _sum_of_weighted_deltas_squared: np.typing.NDArray[Any]

    @property
    def variance(self) -> np.typing.NDArray[Any]:
        with np.errstate(divide="ignore", invalid="ignore"):
            return self["_sum_of_weighted_deltas_squared"] / (
                self["sum_of_weights"]
                - self["sum_of_weights_squared"] / self["sum_of_weights"]
            )


@fields("count", "value", "_sum_of_deltas_squared")
class MeanView(View):
    __slots__ = ()
    _PARENT = Mean

    count: np.typing.NDArray[Any]
    value: np.typing.NDArray[Any]
    _sum_of_deltas_squared: np.typing.NDArray[Any]

    # Variance is a computation
    @property
    def variance(self) -> np.typing.NDArray[Any]:
        with np.errstate(divide="ignore", invalid="ignore"):
            return self["_sum_of_deltas_squared"] / (self["count"] - 1)


def _to_view(
    item: np.typing.NDArray[Any], value: bool = False
) -> np.typing.NDArray[Any] | WeightedSumView | WeightedMeanView | MeanView:
    for cls in View.__subclasses__():
        if item.dtype.names == cls._FIELDS:
            ret = item.view(cls)
            if value and ret.shape:
                return ret.value
            return ret
    return item



from __future__ import annotations

from . import accumulators, axis, numpy, storage
from .histogram import Histogram, IndexingExpr, Kind
from .tag import (  # pylint: disable=redefined-builtin
    loc,
    overflow,
    rebin,
    sum,
    underflow,
)

# pylint: disable-next=import-error
from .version import version as __version__

__all__ = [
    "Histogram",
    "IndexingExpr",
    "Kind",
    "__version__",
    "accumulators",
    "axis",
    "loc",
    "numpy",
    "overflow",
    "rebin",
    "storage",
    "sum",
    "underflow",
]


def __dir__() -> list[str]:
    return __all__



from __future__ import annotations

import copy
from typing import Any, TypeVar

# pylint: disable-next=import-error
from .. import histogram, version
from ._axis import _axis_from_dict, _axis_to_dict
from ._common import serialize_metadata
from ._storage import _data_from_dict, _storage_from_dict, _storage_to_dict

__all__ = ["from_uhi", "remove_writer_info", "to_uhi"]


def __dir__() -> list[str]:
    return __all__


def to_uhi(h: histogram.Histogram, /) -> dict[str, Any]:
    """Convert an Histogram to a dictionary."""

    # Convert the histogram to a dictionary
    data = {
        "uhi_schema": 1,
        "writer_info": {"boost-histogram": {"version": version.version}},
        "axes": [_axis_to_dict(axis) for axis in h.axes],
        "storage": _storage_to_dict(h.storage_type(), h.view(flow=True)),
    }
    data["metadata"] = serialize_metadata(h.__dict__)

    return data


def from_uhi(data: dict[str, Any], /) -> histogram.Histogram:
    """Convert a dictionary to an Histogram."""

    h = histogram.Histogram(
        *(_axis_from_dict(ax) for ax in data["axes"]),
        storage=_storage_from_dict(data["storage"]),
    )
    h[...] = _data_from_dict(data["storage"])
    h.__dict__ = data.get("metadata", {})
    return h


T = TypeVar("T", bound="dict[str, Any]")


def remove_writer_info(obj: T, /, *, library: str | None = "boost-histogram") -> T:
    """
    Removes all ``writer_info`` for a library from a histogram dict, axes dict,
    or storage dict. Makes copies where required, and the outer dictionary is
    always copied.

    Specify a library name, or ``None`` to remove all.
    """

    obj = copy.copy(obj)
    if library is None:
        obj.pop("writer_info")
    elif library in obj.get("writer_info", {}):
        obj["writer_info"] = copy.copy(obj["writer_info"])
        del obj["writer_info"][library]

    if "axes" in obj:
        obj["axes"] = [remove_writer_info(ax, library=library) for ax in obj["axes"]]
    if "storage" in obj:
        obj["storage"] = remove_writer_info(obj["storage"], library=library)

    return obj



from __future__ import annotations

import functools
from typing import Any

from .. import axis
from ._common import serialize_metadata

__all__ = ["_axis_from_dict", "_axis_to_dict"]


def __dir__() -> list[str]:
    return __all__


@functools.singledispatch
def _axis_to_dict(ax: Any, /) -> dict[str, Any]:
    """Convert an axis to a dictionary."""
    raise TypeError(f"Unsupported axis type: {type(ax)}")


@_axis_to_dict.register(axis.Regular)
@_axis_to_dict.register(axis.Integer)
def _(ax: axis.Regular | axis.Integer, /) -> dict[str, Any]:
    """Convert a Regular axis to a dictionary."""

    shared = {
        "underflow": ax.traits.underflow,
        "overflow": ax.traits.overflow,
        "circular": ax.traits.circular,
    }

    # Special handling if the axis has a transform
    if isinstance(ax, axis.Regular) and ax.transform is not None:
        data = {
            "type": "variable",
            "edges": ax.edges,
            **shared,
        }
    elif isinstance(ax, axis.Integer):
        data = {
            "type": "regular",
            "lower": int(ax.edges[0]),
            "upper": int(ax.edges[-1]),
            "bins": ax.size,
            **shared,
        }
    else:
        data = {
            "type": "regular",
            "lower": float(ax.edges[0]),
            "upper": float(ax.edges[-1]),
            "bins": ax.size,
            **shared,
        }

    if isinstance(ax, axis.Integer):
        data["writer_info"] = {"boost-histogram": {"orig_type": "Integer"}}

    data["metadata"] = serialize_metadata(ax.__dict__)

    return data


@_axis_to_dict.register
def _(ax: axis.Variable, /) -> dict[str, Any]:
    """Convert a Variable or Integer axis to a dictionary."""
    data = {
        "type": "variable",
        "edges": ax.edges,
        "underflow": ax.traits.underflow,
        "overflow": ax.traits.overflow,
        "circular": ax.traits.circular,
    }
    data["metadata"] = serialize_metadata(ax.__dict__)

    return data


@_axis_to_dict.register
def _(ax: axis.IntCategory, /) -> dict[str, Any]:
    """Convert an IntCategory axis to a dictionary."""
    data = {
        "type": "category_int",
        "categories": list(ax),
        "flow": ax.traits.overflow,
    }
    data["metadata"] = serialize_metadata(ax.__dict__)

    return data


@_axis_to_dict.register
def _(ax: axis.StrCategory, /) -> dict[str, Any]:
    """Convert a StrCategory axis to a dictionary."""
    data = {
        "type": "category_str",
        "categories": list(ax),
        "flow": ax.traits.overflow,
    }
    data["metadata"] = serialize_metadata(ax.__dict__)

    return data


@_axis_to_dict.register
def _(ax: axis.Boolean, /) -> dict[str, Any]:
    """Convert a Boolean axis to a dictionary."""
    data: dict[str, Any] = {
        "type": "boolean",
    }
    data["metadata"] = serialize_metadata(ax.__dict__)

    return data


def _axis_from_dict(data: dict[str, Any], /) -> axis.Axis:
    orig_type = (
        data.get("writer_info", {}).get("boost-histogram", {}).get("orig_type", "")
    )
    if orig_type == "Integer":
        assert data["upper"] - data["lower"] == data["bins"]
        return axis.Integer(
            data["lower"],
            data["upper"],
            underflow=data["underflow"],
            overflow=data["overflow"],
            circular=data["circular"],
            __dict__=data.get("metadata"),
        )

    hist_type = data["type"]
    if hist_type == "regular":
        return axis.Regular(
            data["bins"],
            data["lower"],
            data["upper"],
            underflow=data["underflow"],
            overflow=data["overflow"],
            circular=data["circular"],
            __dict__=data.get("metadata"),
        )
    if hist_type == "variable":
        return axis.Variable(
            data["edges"],
            underflow=data["underflow"],
            overflow=data["overflow"],
            circular=data["circular"],
            __dict__=data.get("metadata"),
        )
    if hist_type == "category_int":
        return axis.IntCategory(
            data["categories"],
            overflow=data["flow"],
            __dict__=data.get("metadata"),
        )
    if hist_type == "category_str":
        return axis.StrCategory(
            data["categories"],
            overflow=data["flow"],
            __dict__=data.get("metadata"),
        )
    if hist_type == "boolean":
        return axis.Boolean(__dict__=data.get("metadata"))

    raise TypeError(f"Unsupported axis type: {hist_type}")



# bh.sum is just the Python sum, so from boost_histogram import * is safe (but
# not recommended)
from __future__ import annotations

import copy
from builtins import sum
from collections.abc import Sequence
from typing import TYPE_CHECKING, TypeVar

import numpy as np

if TYPE_CHECKING:
    from uhi.typing.plottable import PlottableAxis

from ._compat.typing import Self
from .typing import AxisLike

__all__ = ("Locator", "Slicer", "at", "loc", "overflow", "rebin", "sum", "underflow")


def Slicer() -> np.lib._index_tricks_impl.IndexExpression:
    """
    It is encouraged to use ``np.s_`` directly instead of this function:

        h[{0: np.s_[::bh.rebin(2)]}]   # rebin axis 0 by two

    This is provided for backward compatibility.
    """
    return np.s_


T = TypeVar("T", bound="Locator")


class Locator:
    __slots__ = ("offset",)
    NAME = ""

    def __init__(self, offset: int = 0) -> None:
        if not isinstance(offset, int):
            raise ValueError("The offset must be an integer")

        self.offset = offset

    def __add__(self, offset: int) -> Self:
        other = copy.copy(self)
        other.offset += offset
        return other

    def __sub__(self, offset: int) -> Self:
        other = copy.copy(self)
        other.offset -= offset
        return other

    def _print_self_(self) -> str:
        return ""

    def __repr__(self) -> str:
        s = self.NAME or self.__class__.__name__
        s += self._print_self_()
        if self.offset != 0:
            s += " + " if self.offset > 0 else " - "
            s += str(abs(self.offset))
        return s


class loc(Locator):
    __slots__ = ("value",)

    def __init__(self, value: str | float, offset: int = 0) -> None:
        super().__init__(offset)
        self.value = value

    def _print_self_(self) -> str:
        return f"({self.value})"

    def __call__(self, axis: AxisLike) -> int:
        return axis.index(self.value) + self.offset


class Underflow(Locator):
    __slots__ = ()
    NAME = "underflow"

    def __call__(self, axis: AxisLike) -> int:  # noqa: ARG002
        return -1 + self.offset


underflow = Underflow()


class Overflow(Locator):
    __slots__ = ()
    NAME = "overflow"

    def __call__(self, axis: AxisLike) -> int:
        return len(axis) + self.offset


overflow = Overflow()


class at:
    __slots__ = ("value",)

    def __init__(self, value: int) -> None:
        self.value = value

    def __call__(self, axis: AxisLike) -> int:  # noqa: ARG002
        return self.value


class rebin:
    __slots__ = (
        "axis",
        "edges",
        "factor",
        "groups",
    )

    def __init__(
        self,
        factor_or_axis: int | PlottableAxis | None = None,
        /,
        *,
        factor: int | None = None,
        groups: Sequence[int] | None = None,
        edges: Sequence[int | float] | None = None,
        axis: PlottableAxis | None = None,
    ) -> None:
        if isinstance(factor_or_axis, int):
            factor = factor_or_axis
        elif factor_or_axis is not None:
            axis = factor_or_axis

        total_args = sum(i is not None for i in [factor, groups, edges])
        if total_args != 1 and axis is None:
            raise ValueError("Exactly one argument should be provided")

        self.groups = groups
        self.edges = edges
        self.axis = axis
        self.factor = factor

    def __repr__(self) -> str:
        repr_str = f"{self.__class__.__name__}"
        args: dict[str, int | Sequence[int | float] | PlottableAxis | None] = {
            "factor": self.factor,
            "groups": self.groups,
            "edges": self.edges,
            "axis": self.axis,
        }
        for k, v in args.items():
            if v is not None:
                return_str = f"{repr_str}({k}={v})"
                break
        return return_str

    # Note: this preserves the input type of `self.axis`, so is safe within a
    # single UHI library, but not cross-library. Returns None for the axis if
    # an axis is not provided, the caller should make an axis if that's the
    # case.
    def axis_mapping(
        self, axis: PlottableAxis
    ) -> tuple[Sequence[int], PlottableAxis | None]:
        return (self.group_mapping(axis), self.axis)

    def group_mapping(self, axis: PlottableAxis) -> Sequence[int]:
        if self.groups is not None:
            if sum(self.groups) != len(axis):
                msg = f"The sum of the groups ({sum(self.groups)}) must be equal to the number of bins in the axis ({len(axis)})"
                raise ValueError(msg)
            return self.groups
        if self.factor is not None:
            return [self.factor] * len(axis)
        if self.edges is not None or self.axis is not None:
            newedges = None
            if self.edges is not None:
                newedges = self.edges
            elif self.axis is not None and hasattr(self.axis, "edges"):
                newedges = self.axis.edges

            if newedges is not None and hasattr(axis, "edges"):
                if newedges[0] != axis.edges[0]:
                    msg = "Edges must start at first bin"
                    raise ValueError(msg)
                if newedges[-1] != axis.edges[-1]:
                    msg = "Edges must end at last bin"
                    raise ValueError(msg)
                matched_ixes = [np.abs(axis.edges - edge).argmin() for edge in newedges]
                missing_edges = [
                    edge
                    for ix, edge in zip(matched_ixes, newedges)
                    if not np.isclose(axis.edges[ix], edge)
                ]
                if missing_edges:
                    missing_edges_repr = ", ".join(map(str, missing_edges))
                    msg = f"Edge(s) {missing_edges_repr} not found in axis"
                    raise ValueError(msg)
                return [
                    int(ix - matched_ixes[i]) for i, ix in enumerate(matched_ixes[1:])
                ]
        raise ValueError("No rebinning factor or groups provided")



from __future__ import annotations

from typing import Any


def serialize_metadata(value: dict[str, Any], /) -> dict[str, Any]:
    return {k: v for k, v in value.items() if not k.startswith("@") and v is not None}
